/* MPROM Firmware Code version 1.0(ATAPI Control functions)          */
/* programmed by M.yanagisawa(MoroboshiLUM)  elfin@mth.biglobe.ne.jp */
/* Copyright (c) 2000 by M.yanagisawa. All rigths reserved.          */

#ifdef __BORLANDC__

#include <stdio.h>
#include <dos.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <alloc.h>

#else

#include <stdrom.h>

#endif

#include "mprom.h"
#include "main.h"
#include "atactrl.h"

/* Global code */
BYTE ERROR_STATUS;
int atapi_intr_state;
WORD transmax;
BYTE far *transptr;

#ifdef __BORLANDC__

static void inportdata(WORD far *p,int cnt)
{
   int i;

   for (i=0;i<cnt;i++) *p++=inpw(ATA_DATA);
}


static void outportdata(WORD far *p,int cnt)
{
   int i;

   for (i=0;i<cnt;i++) outpw(ATA_DATA,*p++);
}

#else

void inportdata(WORD far *p,int cnt); /**/
void outportdata(WORD far *p,int cnt); /**/

#endif

/* clear ATA command packet */
void clratapac(PATAPAC pac)
{
  memset(pac,0,sizeof(ATAPAC));
}

/* clear ATAPI command packet */
void clratapipac(PATAPIPAC pac)
{
  memset(pac,0,sizeof(ATAPIPAC));
}

/* wait dataset busy bit */
BOOL wait_data_ready(void)
{
   BYTE stat;
   long i=0;

   do {
     stat=ATA_INP(ATA_ASTAT);
     i++;
   } while ((stat & _BSY) && (i<TIMEOUT_CNT));
   if ((stat & _ERR) || ((stat & _DRQ)==0)) {
     ATA_INP(ATA_STAT);
     return(FALSE);
   }
   return(i<TIMEOUT_CNT);
}

/* wait atapi command ready */
BOOL wait_atapi_packet_ready(void)
{
   BYTE stat,intstat;
   long i=0;

   do {
     stat=ATA_INP(ATA_ASTAT);
     if ((stat & _BSY)==0) {
       if ((stat & _ERR)==0) {
         intstat=ATA_INP(ATA_INTR);
         if ((stat & _DRQ) && 
             ((intstat & _ATAPI_IO)==0) && (intstat & _ATAPI_CD)) break;
       }
     }
     i++;
   } while (!(stat & _ERR) && (i<TIMEOUT_CNT));
   if ((stat & _ERR) || ((stat & _DRQ)==0)) {
     ERROR_STATUS=ATA_INP(ATA_STAT);
     return(FALSE);
   }
   return(i<TIMEOUT_CNT);
}

/* wait atapi in-data ready */
BOOL wait_atapi_indata_ready(WORD *datasize,WORD far *p)
{
   BYTE stat,intstat;
   long i=0;

   stat=ATA_INP(ATA_ASTAT);
   do {
     stat=ATA_INP(ATA_STAT);
     if ((stat & _BSY)==0) {
       if ((stat & _ERR)==0) {
         intstat=ATA_INP(ATA_INTR);
         if ((stat & _DRQ) && 
             (intstat & _ATAPI_IO) && ((intstat & _ATAPI_CD)==0)) {
             *datasize=((WORD)ATA_INP(ATA_BYCH))*256+(WORD)ATA_INP(ATA_BYCL);
             inportdata(p,(*datasize) >> 1);
             break;
          }
       }
     }
     i++;
   } while (!(stat & _ERR) && (i<TIMEOUT_CNT));
   if ((stat & _ERR) || ((stat & _DRQ)==0)) {
     ERROR_STATUS=ATA_INP(ATA_STAT);
     return(FALSE);
   }
   return(i<TIMEOUT_CNT);
}

/* wait atapi out-data ready */
BOOL wait_atapi_outdata_ready(WORD *datasize,WORD far *p)
{
   BYTE stat,intstat;
   long i=0;

   stat=ATA_INP(ATA_ASTAT);
   do {
     stat=ATA_INP(ATA_STAT);
     if ((stat & _BSY)==0) {
       if ((stat & _ERR)==0) {
         intstat=ATA_INP(ATA_INTR);
         if ((stat & _DRQ) && 
             ((intstat & _ATAPI_IO)==0) && ((intstat & _ATAPI_CD)==0)) {
             *datasize=((WORD)ATA_INP(ATA_BYCH))*256+(WORD)ATA_INP(ATA_BYCL);
			 outportdata(p,(*datasize) >> 1);
             break;
          }
       }
     }
     i++;
   } while (!(stat & _ERR) && (i<TIMEOUT_CNT));
   if ((stat & _ERR) || ((stat & _DRQ)==0)) {
     ERROR_STATUS=ATA_INP(ATA_STAT);
     return(FALSE);
   }
   return(i<TIMEOUT_CNT);
}

/* wait atapi command finish */
BOOL wait_atapi_packet_finish(void)
{
   BYTE stat,intstat;
   long i=0;

   stat=ATA_INP(ATA_ASTAT);
   do {
     stat=ATA_INP(ATA_STAT);
     if ((stat & _BSY)==0) {
       if ((stat & _ERR)==0) {
         intstat=ATA_INP(ATA_INTR);
         if (((stat & _DRQ)==0) && 
             (intstat & _ATAPI_IO) && (intstat & _ATAPI_CD)) break;
       }
     }
     i++;
   } while (!(stat & _ERR) && (i<TIMEOUT_CNT));
   return((i<TIMEOUT_CNT) && ((stat & _ERR)==0));
}

/* ATAPI CD-ROM interrupt by INTRQ signal */
void cdrom_intr(void)
{
	BYTE stat,intstat;
	WORD i,transnum,cnt;
	
	stat=ATA_INP(ATA_STAT);
	intstat=ATA_INP(ATA_INTR);
	
	if ( ((intstat & _ATAPI_IO)==0) && (intstat & _ATAPI_CD) && (stat & _DRQ) ) {
	  atapi_intr_state=ATAPI_WAITING_COMMAND;
	} else if ( (intstat & _ATAPI_IO) && ((intstat & _ATAPI_CD)==0) && (stat & _DRQ)) {
	  /* device to host */
	  atapi_intr_state=ATAPI_RECEIVE_DATA;
	  transnum = (ATA_INP(ATA_BYCH) << 8) | ATA_INP(ATA_BYCL);
	  cnt=(transnum >> 1)+(transnum & 1);
	  if (transptr) {
	    inportdata((WORD far *)transptr,cnt);
        transptr+=transnum; /* next address */
	  } else {
	    for (i=0;i<cnt;i++) ATA_INP(ATA_DATA); /* if recvbuf is illegal then dummy read */
	  }
	} else if ( ((intstat & _ATAPI_IO)==0) && ((intstat & _ATAPI_CD)==0) && (stat & _DRQ)) {
	  /* host to device  */
	  atapi_intr_state=ATAPI_SEND_DATA;
	  transnum = (ATA_INP(ATA_BYCH) << 8) | ATA_INP(ATA_BYCL);
	  outportdata((WORD far *)transptr,(transnum >> 1)+(transnum & 1));
      transptr+=transnum; /* next address */
	} else if ( ((intstat & _ATAPI_IO)==0) && ((intstat & _ATAPI_CD)==0) && ((stat & _DRQ)==0) ) {
      atapi_intr_state=ATAPI_BUS_RELEASE;
	} else if ( (intstat & _ATAPI_IO) && (intstat & _ATAPI_CD) && ((stat & _DRQ)==0) ) {
      atapi_intr_state=ATAPI_FINISH;
      outp(_ICU_MASK,inp(_ICU_MASK) | _CDROM_INTR_MASK); /* cdrom intr close !! */
    }
}

/* wait atapi busy bit */
BOOL wait_busy(void)
{
   BYTE stat;
   long i=0;

   do {
     stat=ATA_INP(ATA_STAT);
     i++;
   } while ((stat & _BSY) && ((stat & _ERR)==0) && (i<TIMEOUT_CNT));
   return((!(stat & _ERR)) && (i<TIMEOUT_CNT));
}

void SetCommand(PATAPAC pac)
{
#ifndef __BORLANDC__
   pac->devhead|=0x10; /* Slave Select */
#endif
   ATA_OUTP(ATA_DVCL,0x2);	/**/
   ATA_OUTP(ATA_FERT,pac->feature);
   ATA_OUTP(ATA_SECC,pac->seccnt);
   ATA_OUTP(ATA_SECN,pac->secnum);
   ATA_OUTP(ATA_CLCL,pac->cyllow);
   ATA_OUTP(ATA_CLCH,pac->cylhigh);
   ATA_OUTP(ATA_DVHD,pac->devhead);
   ATA_OUTP(ATA_CMD,pac->cmd);
   inp(FPGAR2);	// 160ns@12MHz
   inp(FPGAR2);	// 160ns@12MHz
   inp(FPGAR2);	// 160ns@12MHz
   inp(FPGAR2);	// 160ns@12MHz
}

void SetCommand2(PATAPAC pac)
{
#ifndef __BORLANDC__
   pac->devhead|=0x10; /* Slave Select */
#endif
   ATA_OUTP(ATA_DVCL,0x0);	/* user intr */
   ATA_OUTP(ATA_FERT,pac->feature);
   ATA_OUTP(ATA_SECC,pac->seccnt);
   ATA_OUTP(ATA_SECN,pac->secnum);
   ATA_OUTP(ATA_CLCL,pac->cyllow);
   ATA_OUTP(ATA_CLCH,pac->cylhigh);
   ATA_OUTP(ATA_DVHD,pac->devhead);
   ATA_OUTP(ATA_CMD,pac->cmd);
   inp(FPGAR2);	// 160ns@12MHz
   inp(FPGAR2);	// 160ns@12MHz
   inp(FPGAR2);	// 160ns@12MHz
   inp(FPGAR2);	// 160ns@12MHz
}

BOOL SendATACommand(PATAPAC pac)
{
   BOOL rslt;

   SetCommand(pac);
   rslt=wait_busy();
   ATA_INP(ATA_STAT);
   return(rslt);
}


BOOL SendATACommand_indata(PATAPAC pac,BYTE far *data,int cnt)
{
   int i;

   SetCommand(pac);
   if (!wait_busy()) {
     ATA_INP(ATA_ASTAT);
     return(FALSE);
   } else {
     for (i=0;i<cnt;i++) {
       if (!wait_data_ready()) return(FALSE);
       inportdata((WORD far *)data,256);
       data+=512;
     }
   }
   ATA_INP(ATA_ASTAT);
   return((ERROR_STATUS=ATA_INP(ATA_STAT)));
}

BOOL SendATACommand_outdata(PATAPAC pac,BYTE far *data,int cnt)
{
   int i;

   SetCommand(pac);
   if (!wait_busy()) {
     ATA_INP(ATA_ASTAT);
     return(FALSE);
   } else {
     for (i=0;i<cnt;i++) {
       if (!wait_data_ready()) return(FALSE);
       outportdata((WORD far *)data,256);
       data+=512;
     }
   }
   ATA_INP(ATA_ASTAT);
   return((ERROR_STATUS=ATA_INP(ATA_STAT)));
}

/* execute ATAPI Command packet without interrupt */
BOOL SendATAPICommand(int _ACCTYPE,ATAPIPAC *apac,int pacsize,BYTE far *p,int cnt)
{
   WORD reqcnt;
   ATAPAC atapipac;

   transmax=cnt;
   clratapac(&atapipac);
   atapipac.cmd=0xa0;	/*ATAPI Packet Command*/
   SetCommand(&atapipac);
   if (!wait_busy()) {
     ATA_INP(ATA_ASTAT);
     return(FALSE);
   } else {
     if (!wait_atapi_packet_ready()) return(FALSE);
     outportdata((WORD far *)apac,pacsize >> 1); /*send packet*/
   }
   switch (_ACCTYPE) {
     case _NODATA  : break;
     case _DATAREAD:
                     if (!wait_atapi_indata_ready(&reqcnt,(WORD far *)p)) return(FALSE);
                     break;
     case _DATAWRITE: 
                     if (!wait_atapi_outdata_ready(&reqcnt,(WORD far *)p)) return(FALSE);
                     break;
   }
   if (!wait_atapi_packet_finish()) return(FALSE);
   ATA_INP(ATA_ASTAT);
   return((ERROR_STATUS=ATA_INP(ATA_STAT)));
}

/* execute ATAPI Command packet with interrupt */
BOOL SendATAPICommand_INTR(ATAPIPAC *apac,int pacsize,BYTE far *p,int cnt)
{
   BOOL IsERR;
   ATAPAC atapipac;
   long timeout;
   BYTE astat;

   IsERR=TRUE;
   atapi_intr_state=ATAPI_WAITING_COMMAND; transptr=p; transmax=cnt;
   clratapac(&atapipac);
   atapipac.cmd=0xa0;	/*ATAPI Packet Command*/
   SetCommand2(&atapipac);
   if (!wait_atapi_packet_ready()) return(FALSE);
   di();
   outp(_ICU_MASK,inp(_ICU_MASK) & ~_CDROM_INTR_MASK); /* enable cdrom intr */
   outportdata((WORD far *)apac,pacsize >> 1);  /*send packet*/
   ei();
   for (timeout=0;timeout<TIMEOUT_CNT;timeout++) {
     astat=ATA_INP(ATA_ASTAT);
     if ((astat & _BSY)==0) {
       if (astat & _ERR) {
         IsERR=TRUE;
         break;
       }
       if (atapi_intr_state==ATAPI_FINISH) {
         IsERR=FALSE;
         break;
       }
     }
   }
   outp(_ICU_MASK,inp(_ICU_MASK) | _CDROM_INTR_MASK); /* cdrom intr close !! */
   ERROR_STATUS=ATA_INP(ATA_STAT);
   return(!IsERR);
}

/* CD-ROM drive Hardware reset! */
BOOL ATAReset(void)
{
   long i;

   ATA_OUTP(ATA_DVCL,0x6);
   DELAY(5);
   ATA_OUTP(ATA_DVCL,0x2);
   DELAY(40);

   for (i=0;i<TIMEOUT_CNT;i++) {
     if ((ATA_INP(ATA_ASTAT) & _BSY)==0) return(TRUE);
     DELAY(1);
   }
   return(FALSE);
}

/* ------------------------------------------------------------------------- */ 

void RequestSense(BYTE far *buf)
{
   ATAPIPAC pac;
   BYTE lbuf[32];
   BYTE far *p=lbuf;

   if (buf!=NULL) p=buf;
   clratapipac(&pac);
   pac.cmd=0x3;
   pac.param[3]=18;
#ifdef __BORLANDC__
   SendATAPICommand(_DATAREAD,&pac,12,p,18);
#else
   SendATAPICommand_INTR(&pac,12,p,18);
#endif
}

BOOL TestUnitReady(void)
{
   ATAPIPAC pac;

   clratapipac(&pac);
   pac.cmd=0x0;
#ifdef __BORLANDC__
   return(SendATAPICommand(_NODATA,&pac,12,NULL,0));
#else
   return(SendATAPICommand_INTR(&pac,12,NULL,0));
#endif
}

BOOL PlayCD(DWORD sta,DWORD len)
{
   ATAPIPAC pac;

   clratapipac(&pac);
   pac.cmd=0xa5;
   pac.param[1]=(sta >> 24) & 0xff;
   pac.param[2]=(sta >> 16) & 0xff;
   pac.param[3]=(sta >> 8) & 0xff;
   pac.param[4]=sta & 0xff;
   pac.param[5]=(len >> 24) & 0xff;
   pac.param[6]=(len >> 16) & 0xff;
   pac.param[7]=(len >> 8) & 0xff;
   pac.param[8]=len & 0xff;
#ifdef __BORLANDC__
   return(SendATAPICommand(_NODATA,&pac,12,NULL,0));
#else
   return(SendATAPICommand_INTR(&pac,12,NULL,0));
#endif
}

int isPlaying(long *sec)
{
   ATAPIPAC pac;
   BYTE buf[32];

   clratapipac(&pac);
   pac.cmd=0x42;
   pac.param[1]=0x40;
   pac.param[2]=0x01;
   pac.param[7]=0x10;
#ifdef __BORLANDC__
   SendATAPICommand(_DATAREAD,&pac,12,buf,16);
#else
   SendATAPICommand_INTR(&pac,12,buf,sizeof(buf));
#endif
   *sec=((DWORD)buf[12] << 24) | ((DWORD)buf[13] << 16) | ((DWORD)buf[14] << 8) | (DWORD)buf[15];
   if ((buf[1] & 0x1f)==0x11) return(buf[6]);
   else return(-1);
}

BOOL PlayStop(void)
{
   ATAPIPAC pac;
   
   clratapipac(&pac);
   pac.cmd=0x4e;
#ifdef __BORLANDC__
   SendATAPICommand(_NODATA,&pac,12,NULL,0);
#else
   SendATAPICommand_INTR(&pac,12,NULL,0);
#endif
   return(TRUE);
}

BOOL isDeviceOK(void)
{
  if (TestUnitReady())
    if (TestUnitReady()) return(TRUE);
  return(FALSE);
}

BOOL SeekSector(DWORD lba)
{
   ATAPIPAC pac;
   BOOL rslt;
 
   clratapipac(&pac);
   pac.cmd=0x2B;
   pac.param[1]=(BYTE)((lba >> 24) & 0xff);
   pac.param[2]=(BYTE)((lba >> 16) & 0xff);
   pac.param[3]=(BYTE)((lba >> 8) & 0xff);
   pac.param[4]=(BYTE)(lba & 0xff);
#ifdef __BORLANDC__
   SendATAPICommand(_NODATA,&pac,12,NULL,0);
#else
   SendATAPICommand_INTR(&pac,12,NULL,0);
#endif
   rslt=((ERROR_STATUS & _ERR)==0);
   if (!rslt) {
     DELAY(1);
     RequestSense(NULL);
   } else
     DELAY(1);
   return(rslt);
}

BOOL ReadSector(DWORD lba,BYTE far *buf,int secnum)
{
   ATAPIPAC pac;
   BOOL rslt;
 
   clratapipac(&pac);
   pac.cmd=0x28;
   pac.param[1]=(BYTE)((lba >> 24) & 0xff);
   pac.param[2]=(BYTE)((lba >> 16) & 0xff);
   pac.param[3]=(BYTE)((lba >> 8) & 0xff);
   pac.param[4]=(BYTE)(lba & 0xff);
   pac.param[6]=0;
   pac.param[7]=(BYTE)(secnum & 0xff);
#ifdef __BORLANDC__
   SendATAPICommand(_DATAREAD,&pac,12,buf,secnum*2048);
#else
   SendATAPICommand_INTR(&pac,12,buf,secnum*2048);
#endif
   rslt=((ERROR_STATUS & _ERR)==0);
   if (!rslt) {
     DELAY(1);
     RequestSense(NULL);
   }
   return(rslt);
}

/* result is read sector number */
/* if result = -1 then error happened! */
/* Over 64KBytes ok! */
int LoadingData(BYTE far *p,DWORD bufsize,DWORD lba,DWORD size)
{
#ifdef __BORLANDC__
#define MAX_READSEC 1 /* max 64k(1segment) per one access */
#else
#define MAX_READSEC 32 /* max 64k(1segment) per one access */
#endif

   DWORD loadsize;
   DWORD ptrn;
   WORD seg,ofs,err;
   WORD rsize,readsecsize,remainsec;
   
   if (bufsize<size) loadsize=bufsize;
   else loadsize=size;

   if (size==0) return(0);
   remainsec=readsecsize=((loadsize-1) >> 11)+1;

   ptrn=(DWORD)p; /* init start address */
   ptrn=((ptrn >> 12) & 0xFFFF0) + (ptrn & 0xFFFF); /* physical address */
   SeekSector(lba);
   while (remainsec>0) {
      /* fix pointer */
     seg=ptrn >> 4; ofs=ptrn & 0xf;
     p=(BYTE far *)MK_FP(seg,ofs); /**/
     if (remainsec<=MAX_READSEC) rsize=remainsec;
     else rsize=MAX_READSEC;
     err=0;
     while ((!ReadSector(lba,p,rsize)) && (err<3)) {
       DELAY(1);
       err++; /* retry max 3 times */
     }
#ifdef __BORLANDC__
	 DELAY(5);
#endif
     if (err>=3) return(0); /* ERROR! */
     ptrn+=rsize*2048;
     lba+=rsize;
     remainsec-=rsize;
   }
   return(readsecsize);
}

void SetFeature(BYTE mode)
{
   ATAPAC pac;

   clratapac(&pac);
   pac.cmd=0x3;
   pac.seccnt=mode;
   SendATACommand(&pac);
}

BOOL GetLastSession(DWORD *lba)
{
   ATAPIPAC pac;
   BOOL rslt;
   BYTE buf[16];
 
   clratapipac(&pac);
   pac.cmd=0x43;
   pac.param[7]=12;
   pac.param[8]=0x40;
#ifdef __BORLANDC__
   SendATAPICommand(_DATAREAD,&pac,12,buf,12);
#else
   SendATAPICommand_INTR(&pac,12,buf,12);
#endif
   *lba=((DWORD)buf[8] << 24) | ((DWORD)buf[9] << 16) | ((DWORD)buf[10] << 8) | (DWORD)buf[11];
   rslt=((ERROR_STATUS & _ERR)==0);
   if (!rslt) {
     DELAY(1);
     RequestSense(NULL);
   }
   return(rslt);
}
